// See USB2LCD.h for version info.

#include "stdafx.h"
#include "USB2LCD.h"

#include "cmds.h"

HANDLE hCom = INVALID_HANDLE_VALUE;

char str[1024];

#define BUF_SIZE	1024
#define MAX_KEY_LEN	256

// Handles any size command or string, n is the length of data
// Currently does no error handling...
#define SENDCMD(n, ...) { \
	DWORD _written = 0;	\
	BYTE _data[n] = __VA_ARGS__;	\
	if (!WriteFile(hCom, _data, n, &_written, NULL)) {	\
	} else if (_written != n) { \
	} \
	FlushFileBuffers(hCom); }

// The one and two byte versions are most common, there is a 'compiled' version of them
static void SendCmd(BYTE a) { SENDCMD(2, {254, a}); }
static void SendCmd(BYTE a, BYTE b) { SENDCMD(3, {254, a, b}); }

/**
Open the COM port (the string must be like \\.\COM#)
Returns INVALID_HANDLE_VALUE if it can't open it (likely doesn't exist or already open)
**/
static HANDLE openCOMPort(wchar_t *port) {
	// Open the port
	DCB comDCB;
	COMMTIMEOUTS comTO;
	HANDLE hCom = CreateFileW(port, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hCom == INVALID_HANDLE_VALUE) { return INVALID_HANDLE_VALUE; }

	// Set the properties of the port (assuming GetCommState/SetCommState don't cause errors)
	comDCB.DCBlength = sizeof(DCB);
	GetCommState(hCom, &comDCB);
	comDCB.BaudRate = CBR_9600;
	comDCB.ByteSize = 8;
	comDCB.StopBits = ONESTOPBIT;
	comDCB.Parity = NOPARITY;
	SetCommState(hCom, &comDCB);

	// Set the timeouts of the port (assuming GetCommTimeouts/SetCommTimeouts don't cause errors)
	GetCommTimeouts(hCom, &comTO);
	comTO.ReadIntervalTimeout = 0;
	comTO.ReadTotalTimeoutMultiplier = 10;
	comTO.ReadTotalTimeoutConstant = 10;
	comTO.WriteTotalTimeoutMultiplier = 10;
	comTO.WriteTotalTimeoutConstant = 100;
	SetCommTimeouts(hCom, &comTO);

	return hCom;
}

/**
Checks to see if a the port is actually available, and if it belongs to the
right service.
**/
static int matchExists(HKEY keyCom, wchar_t *serv, wchar_t *port) {
	DWORD err;
	wchar_t name[BUF_SIZE], data[BUF_SIZE];
	DWORD name_len, data_len;
	DWORD type;
	DWORD j;
	for (j = 0; ; ++j) {
		name_len = BUF_SIZE;
		data_len = BUF_SIZE*sizeof(wchar_t);
		err = RegEnumValue(keyCom, j, name, &name_len, NULL, &type, (LPBYTE)data, &data_len);
		if (err == ERROR_NO_MORE_ITEMS) { return 0; } // not found
		if (err != ERROR_SUCCESS) { return -1; }
		if (_wcsicmp(data, port) == 0) {
			// found the port
			if (_wcsnicmp(name, data, snwprintf(data, BUF_SIZE, L"\\Device\\%s", serv)) == 0) {
				// and it belongs to the right service
				return _wtoi(port+3);
			}
			return 0; // doesn't belong to the right service
		}
	}
}


// +-----------------------+
// ! DISPLAYDLL_DriverName !
// +-----------------------+
//! Returns the driver name.
//! The driver name is used by the driver selection GUI to identify the driver.
//! It's good practice to return a proper name and version for the driver so that
//! it can be clearly identified (or else we would have to rely on the plugin name
//! which is subject to modifications).
USB2LCD_API(LCDS_STR) DISPLAYDLL_DriverName(void) {
	static char plugin_name[] = "USB2LCD Driver v1";
	return plugin_name;
}

// +------------------+
// ! DISPLAYDLL_Usage !
// +------------------+
//! Returns plugin usage hints.
//! The usage text describes the main plugin parameters and how they are assembled
//! in the parameter string (syntax indications).
USB2LCD_API(LCDS_STR) DISPLAYDLL_Usage(void) {
	static char plugin_usage[] = "Either nothing or the hex ID of the device to use";
	return plugin_usage;
}

// +------------------------------+
// ! DISPLAYDLL_DefaultParameters !
// +------------------------------+
//! Returns the plugin default parameters.
//! The default parameters string is used as parameter string after the driver
//! has been initialized.
USB2LCD_API(LCDS_STR) DISPLAYDLL_DefaultParameters(void) {
	static char plugin_default_parameters[] = "";
	return plugin_default_parameters;
}

// +-----------------+
// ! DISPLAYDLL_Init !
// +-----------------+
//! Initializes the plugin.
//! All initialization stuff goes there: opening and configuring IO ports, device detection and probing...
//! Sets ok to 1 if initialization is successful, 0 otherwise.
//! Returns an error message in case of failure, NULL otherwise.
//NOTE: size is not used, the display is automatically set up for all sizes
USB2LCD_API(LCDS_STR) DISPLAYDLL_Init(LCDS_BYTE size_x, LCDS_BYTE size_y, LCDS_STR startup_parameters, LCDS_BOOL *ok) {
	if (hCom != INVALID_HANDLE_VALUE) {
		CloseHandle(hCom);
		hCom = INVALID_HANDLE_VALUE;
	}

	int serial = -1;
	if (strlen(startup_parameters) > 0) {
		serial = strtoul(startup_parameters, NULL, 16);
	}

	DWORD err;
	HKEY keyUsb, keyCom, keyDev, keyParam;

	wchar_t key_name[BUF_SIZE];

	wchar_t id[MAX_KEY_LEN], serv[BUF_SIZE], port[BUF_SIZE];
	DWORD id_len, serv_len, port_len;
	int port_num;

	wchar_t szPort[16];

	DWORD type = 0;
	
	// Get the key containing all of the USB2LCD+ devices
	snwprintf(key_name, BUF_SIZE, L"SYSTEM\\CurrentControlSet\\Enum\\USB\\%s", USB_COM_ID);
	if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, key_name, 0, KEY_ENUMERATE_SUB_KEYS, &keyUsb)) != ERROR_SUCCESS) {
		if (err == ERROR_FILE_NOT_FOUND)
			sprintf(str, "The USB2LCD+ enumeration does not exist, are there any USB2LCD+ devices installed?");
		else
			sprintf(str, "Failed to query the USB2LCD+ enumeration value (0x%lX)", err);
		*ok=0; return str;
	}

	// Get the listing of current serial devices
	if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"HARDWARE\\DEVICEMAP\\SERIALCOMM", 0, KEY_QUERY_VALUE, &keyCom)) != ERROR_SUCCESS) {
		RegCloseKey(keyUsb);
		if (err == ERROR_FILE_NOT_FOUND)
			sprintf(str, "The Device Map Serial COMM registry key does not exist, likely no USB2LCD+ devices are currently attached");
		else
			sprintf(str, "Failed to open the Device Map Serial COMM registry key (0x%lX)", err);
		*ok=0; return str;
	}

	// Go through each device in the list
	bool done = false;
	BYTE buffer[128];
	for (DWORD i = 0; !done; ++i) {
		// Get the next device
		id_len = MAX_KEY_LEN;
		err = RegEnumKeyEx(keyUsb, i, id, &id_len, NULL, NULL, NULL, NULL);
		if (err == ERROR_NO_MORE_ITEMS) { break; }
		if (err) { /*error!*/ break; }

		// Get the service name
		serv_len = BUF_SIZE*sizeof(wchar_t);
		if ((err = RegOpenKeyEx(keyUsb, id, 0, KEY_QUERY_VALUE, &keyDev)) != ERROR_SUCCESS) { /*error!*/ continue; }
		if ((err = RegQueryValueEx(keyDev, L"Service", NULL, &type, (LPBYTE)serv, &serv_len)) != ERROR_SUCCESS || type != REG_SZ) { /*error!*/ RegCloseKey(keyDev); continue; }

		// Get the port name
		port_len = BUF_SIZE;
		if ((err = RegOpenKeyEx(keyDev, L"Device Parameters", 0, KEY_QUERY_VALUE, &keyParam)) != ERROR_SUCCESS) { /*error!*/ RegCloseKey(keyDev); continue; }
		err = RegQueryValueEx(keyParam, L"PortName", NULL, &type, (LPBYTE)port, &port_len);
		
		// Close keys opened so far
		RegCloseKey(keyParam);
		RegCloseKey(keyDev);

		if (err != ERROR_SUCCESS || type != REG_SZ) { /*error!*/ continue; }

		// See if that port is actually active and associated with the service
		port_num = matchExists(keyCom, serv, port);
		if (port_num > 0) {
			swprintf(szPort, L"\\\\.\\COM%u", port_num);

			// Check if this is the right one
			if (serial >= 0) {
				HANDLE hc = openCOMPort(szPort);
				DWORD len = 0;
				while (ReadFile(hc, buffer, 128, &len, NULL) && len == 128); // discard the entire input buffer
				buffer[0] = 254;
				buffer[1] = ReadSerialNum;
				if (WriteFile(hc, buffer, 2, &len, NULL) && len == 2) { // send Read Serial Number command
					FlushFileBuffers(hc);
					DWORD read = 0;
					if (ReadFile(hc, buffer, 2, &len, NULL) && len == 2 && serial == (buffer[0] | buffer[1] << 8)) {
						// Found a matching serial number
						done = true;
					}
				}
				CloseHandle(hc);
			} else {
				// Accept the first one (no serial given)
				done = true;
			}
		} else if (port_num < 0) {
			err = port_num;
			/* error! */
			continue;
		}
	}
	RegCloseKey(keyCom);
	RegCloseKey(keyUsb);

	// Found the port?
	if (serial >= 0 && !done && szPort[0]) { static char error_not_found[] = "A USB2LCD+ device with that ID was not found, did you type the right value and is the device plugged in and properly installed?"; return error_not_found; }
	else if (!szPort[0]) { static char error_not_found2[] = "No USB2LCD+ COM ports found, is the device plugged in and properly installed?"; return error_not_found2; }

	// Double check one last time that we didn't already open the port (this is a fix for the Standby bug)
	if (hCom != INVALID_HANDLE_VALUE) {
		*ok = 1;
		return NULL;
	}

	// Open the port
	hCom = openCOMPort(szPort);
	if (hCom == INVALID_HANDLE_VALUE) {
		static char error_cannot_open[] = "USB2LCD COM port cannot be opened, is another program using it?";
		*ok=0;
		return error_cannot_open;
	}
	*ok=1;
	return NULL;

}

// +------------------+
// ! DISPLAYDLL_Write !
// +------------------+
//! Writes the given string on the display.
//! The string is displayed at the current cursor position (see DISPLAYDLL_SetPosition).
//! It is assumed that the cursor position has been set so that the string can fit.
USB2LCD_API(void) DISPLAYDLL_Write(LCDS_STR str) {
	size_t len = strlen(str);
	for (size_t i = 0; i < len; i++) {
		if (str[i] == 8) { str[i] = 0; } // custom character 0, but now have length saved so null characters are okay
	}
	DWORD written = 0;
	if (!WriteFile(hCom, str, len, &written, NULL)) {
		// handle error
	} else if (written != len) {
		// handle error
	}
	FlushFileBuffers(hCom);
}

// +------------------------+
// ! DISPLAYDLL_SetPosition !
// +------------------------+
//! Moves the display cursor at a given position.
USB2LCD_API(void) DISPLAYDLL_SetPosition(LCDS_BYTE x, LCDS_BYTE y) {
	SENDCMD(4, {254, Position, x, y});
}

// +--------------------------+
// ! DISPLAYDLL_SetBrightness !
// +--------------------------+
USB2LCD_API(void) DISPLAYDLL_SetBrightness(LCDS_BYTE brightness) {
	SendCmd(Backlight, brightness);
}

// +------------------------+
// ! DISPLAYDLL_SetContrast !
// +------------------------+
USB2LCD_API(void) DISPLAYDLL_SetContrast(LCDS_BYTE contrast) {
	SendCmd(Contrast, contrast);
}

// +-------------------------+
// ! DISPLAYDLL_SetBacklight !
// +-------------------------+
USB2LCD_API(void) DISPLAYDLL_SetBacklight(LCDS_BOOL light_on) {
	if (light_on) { SendCmd(DisplayOn, 0); } else { SendCmd(DisplayOff); }
}

// +-----------------------+
// ! DISPLAYDLL_CustomChar !
// +-----------------------+
USB2LCD_API(void) DISPLAYDLL_CustomChar(LCDS_BYTE chr, LCDS_BYTE *data) { // data is size 8, and each byte only has the least 5 bits significant
	SENDCMD(11, {254, DefineCustom, chr-1, data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7]});
}

// +----------------------------+
// ! DISPLAYDLL_CustomCharIndex !
// +----------------------------+
USB2LCD_API(LCDS_BYTE) DISPLAYDLL_CustomCharIndex(LCDS_BYTE index) {
	return index==1 ? 8 : (index-1);
}

// +-----------------+
// ! DISPLAYDLL_Done !
// +-----------------+
USB2LCD_API(void) DISPLAYDLL_Done(void) {
	if (hCom != INVALID_HANDLE_VALUE) {
		CloseHandle(hCom);
		hCom = INVALID_HANDLE_VALUE;
	}
}

// +--------------------+
// ! DISPLAYDLL_ReadKey !
// +--------------------+
// This function attempts to read a keypad attached to the
// LCD. This is a polled function, not an interrupt driven function, so it will
// be up to the DLL to resolve multiple keypresses in a polling interval. Pass
// back a failed keyscan (no key is pressed) by setting the high byte to 0xFF.
// If the MSB is zero, LCD Smartie will assume the keyscan was successful, and
// the data in the low byte will indicate the key. 
USB2LCD_API(LCDS_WORD) DISPLAYDLL_ReadKey(void) {
	BYTE buffer[2];
	DWORD read = 0;
	if (!ReadFile(hCom, buffer, 1, &read, NULL)) {
		// handle error
	}
	return read ? (UINT)buffer[0] : 0xFF00;
}

// +------------------------+
// ! DISPLAYDLL_PowerResume !
// +------------------------+
// Not implemented yet, if it ever is need to add it to the DEF file
/*USB2LCD_API(void) DISPLAYDLL_PowerResume(void) {
}*/

// +-------------------+
// ! DISPLAYDLL_SetGPO !
// +-------------------+
USB2LCD_API(void) DISPLAYDLL_SetGPO(LCDS_BYTE gpo, LCDS_BOOL gpo_on) {
	SendCmd(gpo_on ? GPOon : GPOoff, gpo);
}

// +-------------------+
// ! DISPLAYDLL_SetFan !
// +-------------------+
USB2LCD_API(void) DISPLAYDLL_SetFan(LCDS_BYTE t1, LCDS_BYTE t2) {
	SENDCMD(4, {254, GPOpwm_, t1, t2});
}
